/*
 Copyright (c) MINZ team
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com>
 Made in KOREA

 MIKO server Release 1.0.0b
*/

#ifndef DEF_SOURCE_miko_c
#define DEF_SOURCE_miko_c "miko.c"

#include "miko.h"

int MZ_MIKO(const char *s_ServerName, int s_ServerPort, int s_MaxConnection, int s_ConnectionTimeOut);
static void MZ_MIKO_Disconnect(t_MIKO *s_MIKO);
static int MZ_MIKO_SendHTTP(t_MIKO *s_MIKO, const void *s_Data, int s_Size);
static DEF_COMPLETE_TYPE_BUFFER *MZ_PushBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, const void *s_Data, int s_Length);
static DEF_COMPLETE_TYPE_BUFFER *MZ_DestroyBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle);
static int MZ_PrePopBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, void *s_Data, int s_Length);
static int MZ_PopBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, void *s_Data, int s_Length);
static int MZ_CheckBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle);
static DEF_COMPLETE_TYPE_BUFFER *__MZ_AlignPushBuffer__(DEF_COMPLETE_TYPE_BUFFER *s_Handle);
static void *MZ_MIKO_Alloc(int s_Size);
static void MZ_MIKO_Free(void *s_AllocMemory);

static unsigned long g_MIKO_Memory = (unsigned long)0;

int MZ_MIKO(const char *s_ServerName, int s_ServerPort, int s_MaxConnection, int s_ConnectionTimeOut)
{
 const char * const c_ProtocolName_TCP = "tcp";
 const int          c_BackLog = SOMAXCONN;
 int s_Return = 0, s_ReUseAddress_ON = 1, s_RecvBufferSize = ( 1 << 20 ); 
 struct protoent *s_ProtocolEntry;
 t_MIKO s_MIKO;
 struct hostent *s_HostEntry;
 t_CPU_Tick s_OldTick, s_NewTick;
 char s_ProcBuffer[ 4 << 10 ];
 unsigned long long s_OldSessionCount;
 fprintf(stdout, "MIKO server ready. (port=%d)\n", s_ServerPort);
 s_ProtocolEntry = getprotobyname(c_ProtocolName_TCP); 
 s_MIKO.Socket = socket(AF_INET, SOCK_STREAM, s_ProtocolEntry ? s_ProtocolEntry->p_proto : IPPROTO_IP); 
 if(s_MIKO.Socket >= 0)
 {
  setsockopt(s_MIKO.Socket, SOL_SOCKET, SO_REUSEADDR, &s_ReUseAddress_ON, sizeof(s_ReUseAddress_ON));
  setsockopt(s_MIKO.Socket, SOL_SOCKET, SO_RCVBUF   , &s_RecvBufferSize, sizeof(s_RecvBufferSize)); 
  s_MIKO.BindAddress.sin_family = AF_INET;
  if(s_ServerName)s_HostEntry = gethostbyname(s_ServerName);
  else s_HostEntry = (struct hostent *)0;
  if(s_HostEntry)memcpy((void *)(&s_MIKO.BindAddress.sin_addr), (void *)s_HostEntry->h_addr, sizeof(s_MIKO.BindAddress.sin_addr));
  else s_MIKO.BindAddress.sin_addr.s_addr = htonl(INADDR_ANY);
  s_MIKO.BindAddress.sin_port = htons(s_ServerPort);
  memset((void *)(&s_MIKO.BindAddress.sin_zero[0]), 0, sizeof(s_MIKO.BindAddress.sin_zero));
  s_MIKO.BindAddressLength = (socklen_t)sizeof(s_MIKO.BindAddress);
  if(bind(s_MIKO.Socket, (struct sockaddr *)(&s_MIKO.BindAddress), s_MIKO.BindAddressLength) == 0)
  {
   if(listen(s_MIKO.Socket, c_BackLog) == 0)
   {
    s_MIKO.Quit = s_MIKO.Connection = 0;
    s_MIKO.MaxFD = s_MIKO.Socket + 1;
    s_MIKO.Task = s_MIKO.TraceTask = s_MIKO.NewTask = (t_MIKO_Task *)0;
    s_MIKO.MaxConnection = s_MaxConnection;
    s_MIKO.ConnectionTimeOut = s_ConnectionTimeOut;
    s_MIKO.PeekConnection = 0;
    s_MIKO.TotalRecvBytes = s_MIKO.TotalSendBytes = s_MIKO.SessionCount = (unsigned long long)0;
    memset((void *)(&s_OldTick), 0, sizeof(s_OldTick));
    memset((void *)(&s_NewTick), 0, sizeof(s_OldTick));
    gettimeofday((struct timeval *)(&s_MIKO.OldSelectTime), (struct timezone *)0); /* Check time-stamp */
    s_OldSessionCount = (unsigned long long)0; 
    do
    {
     FD_ZERO(&s_MIKO.FD_RxD);
     FD_ZERO(&s_MIKO.FD_TxD);
     FD_SET(s_MIKO.Socket, &s_MIKO.FD_RxD); /* Retry FD set */
     s_MIKO.MaxFD = s_MIKO.Socket + 1;
     s_MIKO.TraceTask = s_MIKO.Task;
     while(s_MIKO.TraceTask)
     {
      if(s_MIKO.TraceTask->Socket >= 0)
      {
       if(s_MIKO.ConnectionTimeOut < (s_MIKO.SelectTime.tv_sec - s_MIKO.TraceTask->TimeStamp.tv_sec))
       {
        MZ_MIKO_Disconnect((t_MIKO *)(&s_MIKO)); /* Disconnect expire time ! */
       }
       else
       {
        FD_SET(s_MIKO.TraceTask->Socket, &s_MIKO.FD_RxD); /* Retry FD set */
        if(MZ_CheckBuffer(s_MIKO.TraceTask->SendFIFO) > 0)
        {
         FD_SET(s_MIKO.TraceTask->Socket, &s_MIKO.FD_TxD); /* Retry FD set */
        }
        if(s_MIKO.TraceTask->Socket >= s_MIKO.MaxFD)s_MIKO.MaxFD = s_MIKO.TraceTask->Socket + 1; /* Calc new FD count */
       }
      }
      s_MIKO.TraceTask = s_MIKO.TraceTask->Next; 
     }
     s_MIKO.IsSelect = select(s_MIKO.MaxFD, (fd_set *)(&s_MIKO.FD_RxD), (fd_set *)(&s_MIKO.FD_TxD), (fd_set *)0, (struct timeval *)(0));
     gettimeofday((struct timeval *)(&s_MIKO.SelectTime), (struct timezone *)0); /* Check time-stamp */ 
     if(s_MIKO.SelectTime.tv_sec != s_MIKO.OldSelectTime.tv_sec)
     {
      memcpy((void *)(&s_MIKO.OldSelectTime), (void *)(&s_MIKO.SelectTime), sizeof(s_MIKO.OldSelectTime));
      if(MZ_GetCPU_StatBuffer((void *)(&s_ProcBuffer[0]), sizeof(s_ProcBuffer)) > 0)
      {
       if(MZ_GetCPU_Parse(' ', (void *)(&s_ProcBuffer[0]), &s_NewTick) > 0)
       {
        MZ_GetCPU_MakeSubInfo(&s_MIKO.Tick, &s_OldTick, &s_NewTick);    
       }
      }
      s_MIKO.AvgConnection = (int)(s_MIKO.SessionCount - s_OldSessionCount); 
      s_OldSessionCount = s_MIKO.SessionCount;
     }
     if(s_MIKO.IsSelect > 0)
     {
      if(FD_ISSET(s_MIKO.Socket, &s_MIKO.FD_RxD))
      { /* New accept */
       if(s_MIKO.Task)
       {
        s_MIKO.TraceTask = s_MIKO.Task;
        while(s_MIKO.TraceTask)
        {
         if(s_MIKO.TraceTask->Socket < 0)break; /* Keep allocate */
         if(s_MIKO.TraceTask->Next == (t_MIKO_Task *)0)
         { /* New allocate */
          s_MIKO.TraceTask->Next = (t_MIKO_Task *)MZ_MIKO_Alloc(sizeof(t_MIKO_Task)); 
          s_MIKO.TraceTask = s_MIKO.TraceTask->Next;
          s_MIKO.TraceTask->Next = (t_MIKO_Task *)0;
          s_MIKO.TraceTask->RecvFIFO = s_MIKO.TraceTask->SendFIFO = (DEF_COMPLETE_TYPE_BUFFER *)0;
          break;
         }
         s_MIKO.TraceTask = s_MIKO.TraceTask->Next;
        }
        s_MIKO.NewTask = s_MIKO.TraceTask;
       }
       else 
       { /* New allocate : First */
        s_MIKO.NewTask = s_MIKO.Task = (t_MIKO_Task *)MZ_MIKO_Alloc(sizeof(t_MIKO_Task));
        if(s_MIKO.NewTask)
        {
         s_MIKO.NewTask->Next = (t_MIKO_Task *)0;
         s_MIKO.NewTask->RecvFIFO = s_MIKO.NewTask->SendFIFO = (DEF_COMPLETE_TYPE_BUFFER *)0;
        }
       }
       if(s_MIKO.NewTask)
       {
        s_MIKO.NewTask->AcceptAddressLength = sizeof(struct sockaddr_in);
        s_MIKO.NewTask->Socket = accept(s_MIKO.Socket, (struct sockaddr *)(&s_MIKO.NewTask->AcceptAddress), (socklen_t *)(&s_MIKO.NewTask->AcceptAddressLength));
        if(s_MIKO.NewTask->Socket >= 0)
        { /* Accepted new socket */
         if(s_MIKO.MaxConnection == 0 || s_MIKO.Connection < s_MIKO.MaxConnection)
         {
          FD_SET(s_MIKO.NewTask->Socket, &s_MIKO.FD_RxD); /* New FD set */
          s_MIKO.NewTask->TotalRecvBytes = s_MIKO.NewTask->TotalSendBytes = (unsigned long long)0;
          s_MIKO.NewTask->KeepAlive = (-1);
          s_MIKO.Connection++;
          s_MIKO.SessionCount++;
          if(s_MIKO.Connection > s_MIKO.PeekConnection)s_MIKO.PeekConnection = s_MIKO.Connection;
          memcpy((void *)(&s_MIKO.NewTask->TimeStamp), (void *)(&s_MIKO.SelectTime), sizeof(struct timeval));
         }
         else
         { /* Max limited connection !!! */
          close(s_MIKO.NewTask->Socket);
          s_MIKO.NewTask->Socket = (-1); 
         }
        }
       }
      }
      else 
      {
       s_MIKO.TraceTask = s_MIKO.Task;
       while(s_MIKO.TraceTask)
       {
        if(s_MIKO.TraceTask->Socket >= 0)
        {
         if(FD_ISSET(s_MIKO.TraceTask->Socket, &s_MIKO.FD_RxD))
         { /* Can RxD */
          memcpy((void *)(&s_MIKO.TraceTask->TimeStamp), (void *)(&s_MIKO.SelectTime), sizeof(struct timeval));
          s_MIKO.PacketSize = sizeof(s_MIKO.Packet);
          s_MIKO.RecvBytes = recv(s_MIKO.TraceTask->Socket, (void *)(&s_MIKO.Packet[0]), s_MIKO.PacketSize, MSG_NOSIGNAL);
          if(s_MIKO.RecvBytes > 0)
          { /* Recv data */
           s_MIKO.TraceTask->TotalRecvBytes += (unsigned long long)s_MIKO.RecvBytes; 
           s_MIKO.TotalRecvBytes += (unsigned long long)s_MIKO.RecvBytes; 
           s_MIKO.TraceTask->RecvFIFO = MZ_PushBuffer(s_MIKO.TraceTask->RecvFIFO, (void *)(&s_MIKO.Packet[0]), s_MIKO.RecvBytes);
           
           {
            char *s_Title = "\r\n"
                            "<!-- Code by JaeHyuk Cho -->\r\n"
                            "<html>\r\n"
                            "\r\n"
                            "<script language=\"JavaScript\">\r\n"
                            " var g_TimerID = null;\r\n"
                            "\r\n"
                            " function MZ_Go()\r\n"
                            " {\r\n"
                            "  if(g_TimerID)clearTimeout(g_TimerID);\r\n"
                            "  history.go(0);\r\n"
                            " }\r\n"
                            "\r\n"
                            " function MZ_ReDraw()\r\n"
                            " {\r\n"
                            "  g_TimerID = setTimeout(\"MZ_Go()\", %d);\r\n"
                            " }\r\n"
                            "</script>\r\n"
                            "\r\n"
                            " <head>\r\n"
                            "  <title>MIKO test code</title>\r\n"
                            " </head>\r\n"
                            "\r\n"
                            " <body onload=\"MZ_ReDraw()\">\r\n"
                            "  <p>Welcome to MIKO server - Code by JaeHyuk Cho(mailto:minzkn@infoeq.com)</p>\r\n"
                            "  <hr>\r\n"
                            "   <p>Your IP is \"%s:%d\"</p>\r\n"
                            "   <p>MaxFD = %d (Limited 1024)   Descripter </p>\r\n"
                            "   <p>Current connection = %d (MAX = %d) ǽð Ӽ</p>\r\n"
                            "   <p>Connected %d per second (Avg) ʴ  Ӽ</p>\r\n"
                            "   <p>RxD / TxD = %llu / %llu bytes   ۹Ʈ (, TxD ĹǷ 0)</p>\r\n"
                            "   <p>Total RxD / TxD = %llu / %llu bytes   ó </p>\r\n"
                            "   <p>Next time reload %dms (500 + 10ms * Connection))  õ</p>\r\n"
                            "  <hr>\r\n"
                            "   <p>TimeStamp %d/%d/%d %02d:%02d:%02d (SERVER HIT) ð</p>\r\n"
                            "   <p>Session count = %llu  </p>\r\n"
                            "  <hr>\r\n"
                            "   <p>Intel pentium III 1.06GHz mobile CPU (Avg CPU load : %d.%02d%%)</p>\r\n"
                            "   <p>Memory 256MBytes (Used memory : %lu bytes)</p>\r\n"
                            "   <p>Linux kernel 2.4.20</p>\r\n"
                            "   <p>URL: <a href=\"http://www.minzkn.com:2744/\">http://www.minzkn.com:2744/</a></p>\r\n"
                            "   <p>TCP connection style : No thread(Only one process), Standalone, Select(Balance)</p>\r\n"
                            "  <hr>\r\n"
                            " </body>\r\n"
                            "</html>\r\n"
                            "\r\n";
            char s_MakeTitle[ 32 << 10 ];
            int s_NextTime, s_CPU_Load[2];
            time_t s_Time_t;
            struct tm *s_LocalTM;
            time(&s_Time_t);
            s_LocalTM = localtime(&s_Time_t);
            s_NextTime = 500 + (10 * s_MIKO.Connection);
            if(s_MIKO.Tick.Total == 0)s_CPU_Load[0] = s_CPU_Load[1] = 0;
            else
            {
             s_CPU_Load[0] = s_CPU_Load[1] = (int)(((s_MIKO.Tick.Total - s_MIKO.Tick.Idle) * 10000lu) / s_MIKO.Tick.Total);
             s_CPU_Load[0] /= 100;
             s_CPU_Load[1] %= 100;
            }
            sprintf(s_MakeTitle, s_Title, s_NextTime, 
                                 inet_ntoa(s_MIKO.TraceTask->AcceptAddress.sin_addr), 
                                 ntohs(s_MIKO.TraceTask->AcceptAddress.sin_port), 
                                 s_MIKO.MaxFD, 
                                 s_MIKO.Connection, s_MIKO.PeekConnection, s_MIKO.AvgConnection,
                                 s_MIKO.TraceTask->TotalRecvBytes, s_MIKO.TraceTask->TotalSendBytes, 
                                 s_MIKO.TotalRecvBytes, s_MIKO.TotalSendBytes,
                                 s_NextTime,
                                 s_LocalTM->tm_year + 1900, s_LocalTM->tm_mon + 1, s_LocalTM->tm_mday,
                                 s_LocalTM->tm_hour, s_LocalTM->tm_min, s_LocalTM->tm_sec,
                                 s_MIKO.SessionCount++,
                                 s_CPU_Load[0], s_CPU_Load[1],
                                 g_MIKO_Memory
                                );
            MZ_MIKO_SendHTTP((t_MIKO *)(&s_MIKO), (void *)s_MakeTitle, strlen(s_MakeTitle));
            s_MIKO.TraceTask->KeepAlive = 0;
           }
          }
          else MZ_MIKO_Disconnect((t_MIKO *)(&s_MIKO)); /* Disconnected */
          break;
         }
         if(FD_ISSET(s_MIKO.TraceTask->Socket, &s_MIKO.FD_TxD))
         { /* Can TxD */
          memcpy((void *)(&s_MIKO.TraceTask->TimeStamp), (void *)(&s_MIKO.SelectTime), sizeof(struct timeval));
          s_MIKO.PacketSize = MZ_CheckBuffer(s_MIKO.TraceTask->SendFIFO);
          if(s_MIKO.PacketSize > 0)
          {
           if(s_MIKO.PacketSize > sizeof(s_MIKO.Packet))s_MIKO.PacketSize = sizeof(s_MIKO.Packet);
           s_MIKO.PacketSize = MZ_PrePopBuffer(s_MIKO.TraceTask->SendFIFO, (void *)(&s_MIKO.Packet[0]), s_MIKO.PacketSize);  
           s_MIKO.SendBytes = send(s_MIKO.TraceTask->Socket, (void *)(&s_MIKO.Packet[0]), s_MIKO.PacketSize, MSG_NOSIGNAL); 
           if(s_MIKO.SendBytes > 0) /* Send packet */ 
           {
            s_MIKO.TraceTask->TotalSendBytes += (unsigned long long)s_MIKO.SendBytes; 
            s_MIKO.TotalSendBytes += (unsigned long long)s_MIKO.SendBytes; 
            MZ_PopBuffer(s_MIKO.TraceTask->SendFIFO, (void *)(0), s_MIKO.SendBytes); /* Buffer flush */
            if(s_MIKO.TraceTask->KeepAlive == 0)
            {
             if(MZ_CheckBuffer(s_MIKO.TraceTask->SendFIFO) <= 0)MZ_MIKO_Disconnect((t_MIKO *)(&s_MIKO)); /* Disconnected */
            }
           }
           else MZ_MIKO_Disconnect((t_MIKO *)(&s_MIKO)); /* Disconnected */
          }
          break;
         }
        }
        s_MIKO.TraceTask = s_MIKO.TraceTask->Next;
       }
      } /* Is data */
     }
    }while(s_MIKO.Quit == 0);
    while(s_MIKO.Task)
    {
     s_MIKO.TraceTask = s_MIKO.Task;
     s_MIKO.Task = s_MIKO.Task->Next;
     MZ_MIKO_Disconnect((t_MIKO *)(&s_MIKO)); /* Disconnect expire time ! */
     MZ_MIKO_Free(s_MIKO.TraceTask);
    }
   }
  }
  close(s_MIKO.Socket);
 }
 fprintf(stdout, "MIKO server die.\n");
 return(s_Return);
}

static void MZ_MIKO_Disconnect(t_MIKO *s_MIKO)
{
 if(s_MIKO)
 {
  if(s_MIKO->TraceTask)
  {
   if(s_MIKO->TraceTask->Socket >= 0)
   {
    close(s_MIKO->TraceTask->Socket);
    if(s_MIKO->TraceTask->RecvFIFO)s_MIKO->TraceTask->RecvFIFO = MZ_DestroyBuffer(s_MIKO->TraceTask->RecvFIFO);
    if(s_MIKO->TraceTask->SendFIFO)s_MIKO->TraceTask->SendFIFO = MZ_DestroyBuffer(s_MIKO->TraceTask->SendFIFO);
    s_MIKO->TraceTask->Socket = (-1); /* Not use set */
    s_MIKO->Connection--;
   }
  }
 }
}

static int MZ_MIKO_SendHTTP(t_MIKO *s_MIKO, const void *s_Data, int s_Size)
{
 char *s_HTTP = "HTTP/1.0 200 OK\r\n"
                "Server: MIKO\r\n"
                "Content-Type: text/html\r\n"
                "Content-Length: %d\r\n"
                "Connection: close\r\n"
                "\r\n";
 char s_HTTP_Buffer[ 1 << 10 ];
 sprintf(&s_HTTP_Buffer[0], s_HTTP, s_Size); 
 s_MIKO->TraceTask->SendFIFO = MZ_PushBuffer(s_MIKO->TraceTask->SendFIFO, (void *)(&s_HTTP_Buffer[0]), strlen(s_HTTP_Buffer));
 s_MIKO->TraceTask->SendFIFO = MZ_PushBuffer(s_MIKO->TraceTask->SendFIFO, s_Data, s_Size);
 return(MZ_CheckBuffer(s_MIKO->TraceTask->SendFIFO));
}

static DEF_COMPLETE_TYPE_BUFFER *MZ_PushBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, const void *s_Data, int s_Length)
{
 int s_TryPushBytes, s_DataIndex;
 DEF_COMPLETE_TYPE_BUFFER_NODE *s_TraceBufferNode;
 if(s_Handle == (DEF_COMPLETE_TYPE_BUFFER *)0)
 {
  s_Handle = (DEF_COMPLETE_TYPE_BUFFER *)MZ_MIKO_Alloc(sizeof(DEF_COMPLETE_TYPE_BUFFER));
  if(s_Handle)memset((void *)s_Handle, 0, sizeof(DEF_COMPLETE_TYPE_BUFFER));
 }
 if(s_Handle && s_Data && s_Length > 0)
 {
  s_DataIndex = 0; 
  if(s_Handle->AlignIndex == 0 && s_Handle->Head)
  {
   s_TraceBufferNode = s_Handle->Head;
   while(s_TraceBufferNode->Next)s_TraceBufferNode = s_TraceBufferNode->Next;
   s_TryPushBytes = s_TraceBufferNode->Size - s_TraceBufferNode->Current;
   if(s_TryPushBytes > 0)
   {
    if(s_TryPushBytes > s_Length)s_TryPushBytes = s_Length; 
    memcpy(__DEF_PEEK__(s_TraceBufferNode->Data, s_TraceBufferNode->Current), __DEF_PEEK__(s_Data, s_DataIndex), s_TryPushBytes); 
    s_Length                   -= s_TryPushBytes;
    s_DataIndex                += s_TryPushBytes;
    s_TraceBufferNode->Current += s_TryPushBytes;
    s_Handle->TotalSize        += s_TryPushBytes; 
   }
  }
  while(s_Length > 0)
  {
   if(s_Handle->AlignPack == (void *)0)
   {
    s_Handle->AlignPack = (void *)MZ_MIKO_Alloc(__DEF_FIFO_BUFFER_ALIGN__);
    s_Handle->AlignIndex = 0;
    if(s_Handle->AlignPack == (void *)0)break; /* ERROR */
   }
   s_TryPushBytes = __DEF_FIFO_BUFFER_ALIGN__ - s_Handle->AlignIndex;
   if(s_TryPushBytes > s_Length)s_TryPushBytes = s_Length;
   memcpy(__DEF_PEEK__(s_Handle->AlignPack, s_Handle->AlignIndex), __DEF_PEEK__(s_Data, s_DataIndex), s_TryPushBytes); 
   s_Length             -= s_TryPushBytes;
   s_DataIndex          += s_TryPushBytes;
   s_Handle->TotalSize  += s_TryPushBytes; 
   s_Handle->AlignIndex += s_TryPushBytes;
   if(s_Handle->AlignIndex >= __DEF_FIFO_BUFFER_ALIGN__)s_Handle = __MZ_AlignPushBuffer__(s_Handle);
  }
 }
 return(s_Handle);
}

static DEF_COMPLETE_TYPE_BUFFER *MZ_DestroyBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle)
{
 DEF_COMPLETE_TYPE_BUFFER_NODE *s_TraceBufferNode;	
 if(s_Handle)
 {
  if(s_Handle->AlignPack)MZ_MIKO_Free(s_Handle->AlignPack); 
  while(s_Handle->Head)
  {
   s_TraceBufferNode = s_Handle->Head;	  
   s_Handle->Head = s_Handle->Head->Next;
   if(s_TraceBufferNode->Data)MZ_MIKO_Free(s_TraceBufferNode->Data);
   MZ_MIKO_Free(s_TraceBufferNode);
  } 
  MZ_MIKO_Free(s_Handle);	 
  s_Handle = (DEF_COMPLETE_TYPE_BUFFER *)0;
 }
 return(s_Handle);
}

static int MZ_PrePopBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, void *s_Data, int s_Length)
{
 int s_Return = 0, s_TryPopBytes, s_Current;
 DEF_COMPLETE_TYPE_BUFFER_NODE *s_Head;
 if(s_Handle && s_Length > 0)
 {
  if(s_Handle->AlignPack && s_Handle->AlignIndex > 0)s_Handle = __MZ_AlignPushBuffer__(s_Handle); 
  if(s_Handle->Head)
  {
   s_Head = s_Handle->Head;
   s_Current = s_Head->Current;
   while(s_Length > 0 && s_Head /* REAL */)
   {
    s_TryPopBytes = s_Head->Size - s_Current;	  
    if(s_TryPopBytes >= s_Length)s_TryPopBytes = s_Length;	  
    if(s_Data)memcpy(__DEF_PEEK__(s_Data, s_Return), __DEF_PEEK__(s_Head->Data, s_Current), s_TryPopBytes);
    s_Length -= s_TryPopBytes;
    s_Return += s_TryPopBytes;
    s_Current += s_TryPopBytes;
    if(s_Current >= s_Head->Size)
    {
     s_Head = s_Head->Next;
     if(s_Head)s_Current = s_Head->Current;
    }
   }
  }
 }
 return(s_Return);
}

static int MZ_PopBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle, void *s_Data, int s_Length)
{
 int s_Return = 0, s_TryPopBytes;
 DEF_COMPLETE_TYPE_BUFFER_NODE *s_TraceBufferNode;
 if(s_Handle && s_Length > 0)
 {
  if(s_Handle->AlignPack && s_Handle->AlignIndex > 0)s_Handle = __MZ_AlignPushBuffer__(s_Handle); 
  while(s_Length > 0 && s_Handle->Head /* REAL */)
  {
   s_TryPopBytes = s_Handle->Head->Size - s_Handle->Head->Current;	  
   if(s_TryPopBytes >= s_Length)s_TryPopBytes = s_Length;	  
   if(s_Data)memcpy(__DEF_PEEK__(s_Data, s_Return), __DEF_PEEK__(s_Handle->Head->Data, s_Handle->Head->Current), s_TryPopBytes);
   s_Length -= s_TryPopBytes;
   s_Return += s_TryPopBytes;
   s_Handle->Head->Current += s_TryPopBytes;
   if(s_Handle->Head->Current >= s_Handle->Head->Size)
   {
    s_TraceBufferNode = s_Handle->Head;	   
    s_Handle->Head = s_Handle->Head->Next;
    if(s_TraceBufferNode->Data)MZ_MIKO_Free(s_TraceBufferNode->Data);
    MZ_MIKO_Free(s_TraceBufferNode);
   }
  }
  s_Handle->TotalSize -= s_Return; /* REAL */ 
 }
 return(s_Return);
}

static int MZ_CheckBuffer(DEF_COMPLETE_TYPE_BUFFER *s_Handle)
{
 int s_Return;
 if(s_Handle)s_Return = s_Handle->TotalSize;
 else s_Return = 0; 
 return(s_Return);
}

static DEF_COMPLETE_TYPE_BUFFER *__MZ_AlignPushBuffer__(DEF_COMPLETE_TYPE_BUFFER *s_Handle)
{
 DEF_COMPLETE_TYPE_BUFFER_NODE *s_BufferNode, *s_TraceBufferNode;
 s_BufferNode = (DEF_COMPLETE_TYPE_BUFFER_NODE *)MZ_MIKO_Alloc(sizeof(DEF_COMPLETE_TYPE_BUFFER_NODE));
 if(s_BufferNode)
 {
  s_BufferNode->Next    = (DEF_COMPLETE_TYPE_BUFFER_NODE *)0;
  s_BufferNode->Size    = s_Handle->AlignIndex;
  s_BufferNode->Current = 0;
  s_BufferNode->Data    = s_Handle->AlignPack;
  s_Handle->AlignPack   = (void *)0;
  s_Handle->AlignIndex  = 0;
  if(s_Handle->Head)
  {
   s_TraceBufferNode = s_Handle->Head;
   while(s_TraceBufferNode->Next)s_TraceBufferNode = s_TraceBufferNode->Next;
   s_TraceBufferNode->Next = s_BufferNode;
  }
  else s_Handle->Head = s_BufferNode;
 }
 return(s_Handle);
}

static void *MZ_MIKO_Alloc(int s_Size)
{
 void *s_Return;
 if(s_Size > 0)
 {
  s_Return = malloc(sizeof(int) + s_Size); 
  if(s_Return)
  {
   g_MIKO_Memory += s_Size;
   *((int *)s_Return) = s_Size;
   s_Return = (void *)(((int *)s_Return) + 1);
   return(s_Return);
  }
 }
 return((void *)0);
}

static void MZ_MIKO_Free(void *s_AllocMemory)
{
 if(s_AllocMemory)
 { 
  s_AllocMemory = (void *)(((int *)s_AllocMemory) - 1);
  g_MIKO_Memory -= *((int *)s_AllocMemory);
  free(s_AllocMemory);
 }
}

#endif

/* End of source */
