/*
 Copyright (c) MINZ team
 Code by JaeHyuk Cho <mailto:minzkn@infoeq.com>
 Made in KOREA
*/

#ifndef DEF_SOURCE_miko_h
#define DEF_SOURCE_miko_h "miko.c"

/* Normal include */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <time.h>
#include <sched.h>

/* Network include */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>

#include "cpuusage.h"

#define __DEF_FIFO_BUFFER_ALIGN__           (32 << 10)
#define __DEF_PEEK__(s_Data,s_Offset)       ((void *)(((unsigned char *)(s_Data)) + (s_Offset)))

#define DEF_MIKO_PACKET_SIZE                (32 << 10)

typedef struct ts_MZ_BufferNode
{
 struct ts_MZ_BufferNode       *Next;
 int                            Size, Current;
 void                          *Data;
}t_MZ_BufferNode;
#define DEF_COMPLETE_TYPE_BUFFER_NODE t_MZ_BufferNode

typedef struct 
{
 int                            Lock, AlignIndex, TotalSize;
 void                          *AlignPack; 
 DEF_COMPLETE_TYPE_BUFFER_NODE *Head;
}t_MZ_Buffer;
#define DEF_COMPLETE_TYPE_BUFFER      t_MZ_Buffer

typedef struct ts_MIKO_Task
{
 struct ts_MIKO_Task      *Next;
 int                       Socket, KeepAlive;
 unsigned long long        TotalRecvBytes, TotalSendBytes;
 struct sockaddr_in        AcceptAddress;
 socklen_t                 AcceptAddressLength;
 DEF_COMPLETE_TYPE_BUFFER *RecvFIFO, *SendFIFO; 
 struct timeval            TimeStamp;
}t_MIKO_Task;

typedef struct ts_MIKO
{
 int                       Socket, MaxConnection, Quit, MaxFD, Connection, ConnectionTimeOut, IsSelect, PeekConnection;
 struct sockaddr_in        BindAddress;
 socklen_t                 BindAddressLength;
 fd_set                    FD_RxD, FD_TxD;
 t_MIKO_Task              *Task, *TraceTask, *NewTask;
 unsigned char             Packet[ DEF_MIKO_PACKET_SIZE ];
 int                       PacketSize, RecvBytes, SendBytes;
 struct timeval            SelectTime, OldSelectTime;
 unsigned long long        TotalRecvBytes, TotalSendBytes, SessionCount;
 int                       AvgConnection; /* x/second */
 t_CPU_Tick                Tick;
}t_MIKO;

#ifndef DEF_SOURCE_miko_c

extern int MZ_MIKO(const char *s_ServerName, int s_ServerPort, int s_MaxConnection, int s_ConnectionTimeOut);

#endif

#endif

/* End of source */
